/*
QBL.h
Written by Walter O. Krawec

Copyright (c) 2013 Walter O. Krawec

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef QBL_H_
#define QBL_H_

#define QBL_VERSION "0.81a.4.13"

#include "QDM/QDM.h"

#include "WOK/Interpreter.h"
#include <vector>

namespace qbl
{
	class QBLCommands;
	class DistributionType : public wok::interp::UserObject
	{
		friend class QBLCommands;
	public:
		DistributionType () : wok::interp::UserObject("qbl::Distribution") {};
		virtual ~DistributionType() {};

		virtual void print(std::ostream& f)
		{
			if(p.empty())
				f << "EMPTY";
			else
			{
				f << "\n";
				for(int i=0; i<(int)p.size(); ++i)
				{
					f << "(";
					for(int j=0; j<(int)p[i].actionList.size(); ++j)
						f << "[" << p[i].actionList[j].actionSpaceIndex << ";" << p[i].actionList[j].actionIndex << "], ";
					f << ") = " << p[i].p << "\n";
				}
					//f << "(" << p[i].actionSpace << ", " << p[i].action << ") = " << p[i].p << "\n";
			}
		}

		virtual void copy(wok::interp::UserObject* o)
		{
			if(!o || o->getType() != getType())
				return;

			p = ((DistributionType*)o)->p;
		}

	private:
		std::vector <qdm::TrainingDistribution> p;
	};

	class SensorListType : public wok::interp::UserObject
	{
		friend class QBLCommands;
	public:
		SensorListType() : wok::interp::UserObject("qbl::SensorList") {};

		virtual void print(std::ostream& f)
		{
			if(v.empty()) f << "EMPTY";
			else
			{
				for(int i=0; i<(int)v.size(); ++i)
					f << v[i] << ", ";
			}
		}

		virtual void copy(wok::interp::UserObject* o)
		{
			if(!o || o->getType() != getType())
				return;

			v = ((SensorListType*)o)->v;
		}

	private:
		std::vector < std::string > v;
	};

	class StateType : public wok::interp::UserObject
	{
		friend class QBLCommands;
	public:
		StateType() : wok::interp::UserObject("qbl::State") {};

		virtual void print(std::ostream& f)
		{
			f << "\nMode = " << state.mode << "\n";
			for(int i=0; i<(int)state.sensors.size(); ++i)
			{
				f << i << ": ";
				for(int j=0; j<(int)state.sensors[i].size(); ++j)
					f << state.sensors[i][j] << ", ";
				f << "\n";
			}
		}

		virtual void copy(wok::interp::UserObject* o)
		{
			if(!o || o->getType() != getType())
				return;

			state = ((StateType*)o)->state;
		}

	private:
		qdm::TrainingSensorSet state;
	};

	class TrainingSetType : public wok::interp::UserObject
	{
		friend class QBLCommands;
	public:
		TrainingSetType() : wok::interp::UserObject("qbl::TrainingSet") {};

		virtual void print(std::ostream& f)
		{
			if(set.empty()) f << "EMPTY";
			else
			{
				f << "Number of Rules = " << set.size() << "\n";
			}
		}

		virtual void copy(wok::interp::UserObject* o)
		{
			if(!o || o->getType() != getType())
				return;

			set = ((TrainingSetType*)o)->set;
		}

	private:
		std::vector < qdm::TrainingSet > set;
	};

	class QDMType : public wok::interp::UserObject
	{
		friend class QBLCommands;
	public:
		QDMType(qdm::QDM* _qdm) : wok::interp::UserObject("qbl::QDM"), qdm(_qdm) {};
		virtual ~QDMType() {}

		virtual void copy(wok::interp::UserObject* o)
		{
			if(!o || o->getType() != getType())
				return;

			qdm = ((QDMType*)o)->qdm;
		}

	private:
		qdm::QDM* qdm;
	};

	class ActionListType;

	class PointType : public wok::interp::UserObject
	{
		friend class QBLCommands;
	public:
		PointType() : wok::interp::UserObject("qbl::qdm::Point") {};
		virtual ~PointType() {};

		virtual void print(std::ostream& f)
		{
			f << "Point = (";
			for(int i=0; i<(int)p.coord.size(); ++i)
				f << p.coord[i] << ", ";
			f << ")";
		}

		virtual void copy(wok::interp::UserObject* o)
		{
			if(!o || o->getType() != getType())
				return;

			p = ((PointType*)o)->p;
		}
	private:
		qdm::Point p;
	};

	class ActionType : public wok::interp::UserObject
	{
		friend class QBLCommands;
		friend class ActionListType;
	public:
		ActionType() : wok::interp::UserObject("qbl::qdm::Action") {};
		virtual ~ActionType() {};

		virtual void print(std::ostream& f)
		{
			f << "Action = (" << a.name;
			for(int i=0; i<(int)a.p.coord.size(); ++i)
				f << a.p.coord[i] << ", ";
			f << ")";
		}

		virtual void copy(wok::interp::UserObject* o)
		{
			if(!o || o->getType() != getType())
				return;

			a = ((ActionType*)o)->a;
		}

		virtual void add(wok::interp::Object& A, wok::interp::Object& B, wok::interp::Object& output);
	private:
		//qdm::Point p;
		qdm::Action a;
	};

	class ActionListType : public wok::interp::UserObject
	{
		friend class QBLCommands;
		friend class ActionType;
	public:
		ActionListType() : wok::interp::UserObject("qbl::qdm::ActionList") {};
		virtual ~ActionListType() {};

		virtual void print(std::ostream& f)
		{
			for(int j=0; j<(int)actions.size(); ++j)
			{
				f << "Action Point " << j << " = (" << actions[j].name << ", ";
				for(int i=0; i<(int)actions[j].p.coord.size(); ++i)
					f << actions[j].p.coord[i] << ", ";
				f << ")\n";
			}
		}

		virtual void copy(wok::interp::UserObject* o)
		{
			if(!o || o->getType() != getType())
				return;

			actions = ((ActionListType*)o)->actions;
		}

		virtual void add(wok::interp::Object& A, wok::interp::Object& B, wok::interp::Object& output);
	private:
		std::vector <qdm::Action> actions;
	};

	class ActionSpaceType : public wok::interp::UserObject
	{
		friend class QBLCommands;
	public:
		ActionSpaceType() : wok::interp::UserObject("qbl::qdm::ActionSpace") {};
		virtual ~ActionSpaceType() {};

		virtual void print(std::ostream& f)
		{
			f << "\nA.gridSize = " << A.gridSize << "\n";
			f << "A.movementMap = [\n";
			for(int i=0; i<(int)A.movementMap.size(); ++i)
			{
				f << "(";
				for(int j=0; j<(int)A.movementMap[i].coord.size(); ++j)
					f << A.movementMap[i].coord[j] << ", ";
				f << ")\n";
			}
			f << "]\n";
			f << "A.points = ";
			if(A.actions.empty())
				f << "EMPTY\n";
			else
				f << "\n";
			for(int i=0; i<(int)A.actions.size(); ++i)
			{
				f << "\t(" << A.actions[i].name << ", ";
				for(int j=0; j<(int)A.actions[i].p.coord.size(); ++j)
					f << A.actions[i].p.coord[j] << ", ";
				f << ")\n";
			}
		}

		virtual void copy(wok::interp::UserObject* o)
		{
			if(!o || o->getType() != getType())
				return;

			A = ((ActionSpaceType*)o)->A;
		}
	private:
		qdm::ActionSpace A;
	};

	class QBLCommands : public wok::interp::CommandLibrary
	{
	public:
		virtual int requiredVersion() {return 1;};
		virtual bool evaluate(wok::Text& text, wok::interp::Store& store, wok::interp::Object& output, wok::interp::Interpreter* interpreter, std::vector <wok::interp::CommandLibrary*> &library, bool globalMode);

		virtual bool newUserType(const std::string& type, const std::string& name, wok::interp::Store* store)
		{
			if(type == "qbl::dist")
			{
				DistributionType *t = new DistributionType;
				wok::interp::Object o(name, wok::interp::UserType, t);
				store->addStore(o);
			}
			else if(type == "qbl::SensorList")
			{
				SensorListType *t = new SensorListType;
				wok::interp::Object o(name, wok::interp::UserType, t);
				store->addStore(o);
			}
			else if(type == "qbl::State")
			{
				StateType *t = new StateType;
				wok::interp::Object o(name, wok::interp::UserType, t);
				store->addStore(o);
			}
			else if(type == "qbl::TrainingSet")
			{
				TrainingSetType *t = new TrainingSetType;
				wok::interp::Object o(name, wok::interp::UserType, t);
				store->addStore(o);
			}
			else if(type == "qbl::QDM")
			{
				QDMType *t = new QDMType(NULL);
				wok::interp::Object o(name, wok::interp::UserType, t);
				store->addStore(o);
			}
			else if(type == "qbl::qdm::Point")
			{
				PointType *t = new PointType;
				wok::interp::Object o(name, wok::interp::UserType, t);
				store->addStore(o);
			}
			else if(type == "qbl::qdm::Action")
			{
				ActionType *t = new ActionType;
				wok::interp::Object o(name, wok::interp::UserType, t);
				store->addStore(o);
			}
			else if(type == "qbl::qdm::ActionList")
			{
				ActionListType *t = new ActionListType;
				wok::interp::Object o(name, wok::interp::UserType, t);
				store->addStore(o);
			}
			else if(type == "qbl::qdm::ActionSpace")
			{
				ActionSpaceType *t = new ActionSpaceType;
				wok::interp::Object o(name, wok::interp::UserType, t);
				store->addStore(o);
			}
			else
				return false;

			return true;
		}

#ifdef USE_VISUAL_DEBUG
		void setGraphClient(wok::NetRTGraphClient* _graphClient)
		{
			graphClient = _graphClient;
		}
#endif
	private:
#ifdef USE_VISUAL_DEBUG
		wok::NetRTGraphClient* graphClient;
#endif
	};

	class QBL
	{
	public:
		QBL()
		{
#ifdef USE_VISUAL_DEBUG
			graphClient = NULL;
#endif
		}
		void shell(std::vector <wok::interp::CommandLibrary*>* extras=NULL)
		{
#ifdef USE_VISUAL_DEBUG
			qblCommands.setGraphClient(graphClient);
#endif

			std::vector <wok::interp::CommandLibrary*> library;
			library.push_back(&qblCommands);
			if(extras)
			{
				for(int i=0; i<(int)extras->size(); ++i)
					library.push_back((*extras)[i]);
			}
			interpreter.shell("QBL", library);
		}
#ifdef USE_VISUAL_DEBUG
		void setGraphClient(wok::NetRTGraphClient* _graphClient)
		{
			graphClient = _graphClient;
		}
#endif

		void setup(qdm::QDM* qdm)
		{
			interpreter.getGlobalStore().cleanup();

			QDMType *t = new QDMType(qdm);
			wok::interp::Object o("QDM", wok::interp::UserType, t);
			interpreter.getGlobalStore().addStore(o);
		}

		void evaluate(std::string line, std::vector <wok::interp::CommandLibrary*>* extras=NULL)
		{
			std::vector <wok::interp::CommandLibrary*> library;
			library.push_back(&qblCommands);
			if(extras)
			{
				for(int i=0; i<(int)extras->size(); ++i)
					library.push_back((*extras)[i]);
			}
			wok::interp::Object output;
			interpreter.evaluate(line, interpreter.getGlobalStore(), output, library);
		}

		void evaluate(std::string line, wok::interp::Object& output, std::vector <wok::interp::CommandLibrary*>* extras=NULL)
		{
			std::vector <wok::interp::CommandLibrary*> library;
			library.push_back(&qblCommands);
			if(extras)
			{
				for(int i=0; i<(int)extras->size(); ++i)
					library.push_back((*extras)[i]);
			}
			interpreter.evaluate(line, interpreter.getGlobalStore(), output, library);
		}
	private:
		wok::interp::Interpreter interpreter;
		QBLCommands qblCommands;

#ifdef USE_VISUAL_DEBUG
		wok::NetRTGraphClient* graphClient;
#endif
	};
};

#endif