/*
usb-maestro.h
Written by Walter O. Krawec

Copyright (c) 2013 Walter O. Krawec

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

/*
This program is a simple class to drive a Pololu Maestro USB motor controler
We add this functionality to QBL by createing a Command Library (see maestro-interp.h)
*/

#ifndef WOK_ROBOTS_USB_MAESTRO_H_
#define WOK_ROBOTS_USB_MAESTRO_H_

#include <iostream>
#include <fcntl.h>
#include <termios.h>

namespace wok
{
namespace robots
{

class Maestro
{
 public:
 Maestro() : fd(-1) {};
  bool connect(const char* path)
  {
    termios tio;
    tio.c_iflag = 0;
    tio.c_oflag=0;
    tio.c_cflag = CS8|CREAD|CLOCAL;
    tio.c_lflag = 0;
    tio.c_cc[VMIN] = 1;
    tio.c_cc[VTIME] = 5;

    if(!path)
      fd = open("/dev/ttyACM0", O_RDWR);
    else
      fd = open(path, O_RDWR);

    if(fd == -1) return false;
    cfsetospeed(&tio, B9600);
    cfsetispeed(&tio, B9600);
    tcsetattr(fd, TCSANOW, &tio);

    return true;
  }

  void disconnect()
  {
    if(fd == -1) return;
    close(fd);
  }

  // target should be in range [0,1]
  bool moveServo(unsigned char channel, float target)
  {
    if(fd == -1) return false;
    target *= 3000*4;
    int targetInt = (int)target;  // round to nearest int
    unsigned char packet[] = {0x84, channel, targetInt & 0x7f, (targetInt >> 7) & 0x7f};
    if(write(fd, packet, 4) == -1) return false;
    return true;
  }

  int getAnalogInput(unsigned char channel)
  {
    if(fd == -1) return false;
    unsigned char packet[] = {0x90, channel};
    if(write(fd, packet, 2) == -1) return 0;

    unsigned char byte1, byte2;
    if(read(fd, &byte1, 1) == -1) return 0;
    if(read(fd, &byte2, 1) == -1) return 0;

    return byte1 + 256*byte2;
  }
 private:
  int fd;
};

}
}

#endif
